# frozen_string_literal: true

module Dwarves
  module Parser
    class AST; end

    # The basic DWARF descriptive unit, the Debugging Information Entry (DIE).
    #
    # This is a node in the parse tree that DWARF generates.
    #
    # These represent types, variables, and functions.
    #
    class DIE < AST
      attr_reader *%i[
        tag
        children
        siblings
        attributes
        address
        depth
      ]

      def initialize opts = {}
        @tag = opts[:tag]
        @children = opts[:children] || []
        @siblings = opts[:siblings] || []
        @attributes = opts[:attributes] || []
        @address = opts[:address]
        @depth = opts[:depth]
        @abbrev_number = opts[:abbrev_number]
      end

      def to_s
        attrs = if @attributes.empty?
                  ""
                else
                  %(\n#{@attributes.map(&:to_s).join("\n")})
                end
        depth_addr =
          "<#{@depth}><#{@address}>: Abbrev Number: #{@abbrev_number}"

        if @tag
          " #{depth_addr} (#{@tag})#{attrs}"
        else
          " #{depth_addr}#{attrs}"
        end
      end
    end

    # Compilation Unit
    #
    # @todo clean this up
    class CU < AST
      attr_reader *%i[
        compilation_unit_offset
        length
        architecture
        version
        abbrev_offset
        pointer_size
        dies
      ]

      def initialize opts = {}
        @compilation_unit_offset = opts[:compilation_unit_offset]
        @length = opts[:length]
        @architecture = opts[:architecture]
        @version = opts[:version]
        @abbrev_offset = opts[:abbrev_offset]
        @pointer_size = opts[:pointer_size]
        @dies = opts[:dies] || []
      end

      def to_s
        "Contents of the .debug_info section:\n" \
        "\n" \
        "  Compilation Unit @ offset #{@compilation_unit_offset}:\n" \
        "   Length:        #{@length} (#{@architecture})\n" \
        "   Version:       #{@version}\n" \
        "   Abbrev Offset: #{@abbrev_offset}\n" \
        "   Pointer Size:  #{@pointer_size}\n" \
        "#{@dies.map(&:to_s).join("\n")}\n\n"
      end
    end

    class Attribute < AST
      attr_reader :name, :value, :address

      def initialize opts = {}
        @name = opts[:name]
        @value = opts[:value]
        @address = opts[:address]
      end

      def to_s
        formatted_address = "<#{@address}>".ljust(3, " ")
        formatted_name = name.to_s.ljust(18, " ") + ":"
        "    #{formatted_address}   #{formatted_name} #{@value}"
      end
    end
  end
end
