# frozen_string_literal: true

require "parslet"

module Dwarves
  module Parser
    class V4Parser < Parslet::Parser
      include Parser::BasicRules

      # @!method root
      root :input

      # @!method input
      #
      # The entire input
      rule :input do
        prelude >> die.repeat.as(:dies) >> ending
      end

      # @!method prelude_header
      #
      # The very beginning of input.
      #
      # @example
      #   contents of the .debug_info section:
      #
      # @example
      #   tmp/everything.so:     file format elf64-x86-64
      #
      #   contents of the .debug_info section:
      #
      rule :prelude_header do
        readelf =
          str("Contents of the .debug_info section:") >>
          newline >>
          _

        objdump =
          filename >> _ >> str("file format") >> _ >> format >>
          newline >>
          _ >> str("Contents of the .debug_info section:") >>
          newline >>
          _

        readelf | objdump
      end

      # @!method prelude
      #
      # @example
      #
      #   contents of the .debug_info section:
      #
      #     Compilation Unit @ offset 0x0:
      #      Length:        0xf8 (32-bit)
      #      Version:       4
      #      Abbrev Offset: 0x0
      #      Pointer Size:  8
      #
      # @todo convert numeric string results to numeric values
      #
      rule :prelude do
        prelude_header >>
          str("Compilation Unit @ offset ") >>
          hexadecimal.as(:compilation_unit_offset) >> colon >>
          newline >>
          _ >> str("Length:") >> _ >> hexadecimal.as(:length) >> _ >>
          str("(") >> arch.as(:architecture) >> str(")") >>
          newline >>
          _ >> str("Version:") >> _ >> integer.as(:version) >>
          newline >>
          _ >> str("Abbrev Offset:") >> _ >> hexadecimal.as(:abbrev_offset) >>
          newline >>
          _ >> str("Pointer Size:") >> _ >> integer.as(:pointer_size) >>
          newline
      end

      # @!method arch
      #
      # @example
      #   32-bit
      #
      # @example
      #   64-bit
      #
      rule(:arch) { digit.repeat >> str("-bit") }

      # @!method format
      #
      # @example
      #   elf64-x86-64
      #
      rule(:format) { filename }

      # @!method die_header
      #
      #  @example
      #    <0><b>: Abbrev Number: 1 (DW_TAG_compile_unit)
      #
      #  @example
      #    <3><b8>: Abbrev Number: 0
      #
      rule :die_header do
        _ >>
          str("<") >>
          integer.as(:depth) >>
          str(">") >>
          str("<") >>
          word.as(:address) >>
          str(">") >>
          str(": Abbrev Number: ") >>
          integer.as(:abbrev_number) >>
          (
            _ >>
            str("(") >>
            words.as(:type) >>
            str(")")
          ).maybe >>
          newline
      end

      # @!method die_body
      #
      # @example
      #   <1><2d>: Abbrev Number: 2 (abbrev_number)
      #      <2e>   DW_AT_byte_size   : 4
      #      <2f>   DW_AT_encoding    : 5  (signed)
      #      <30>   DW_AT_name        : int
      #
      rule(:die_body) { attribute.repeat.as :attributes }

      # @!method attribute
      #
      # @example
      #   <30>   DW_AT_name        : int
      #
      rule :attribute do
        _ >>
          str("<") >>
          word.as(:address) >>
          str(">") >>
          _ >>
          word.as(:name) >>
          _ >>
          colon >>
          _ >>
          (newline.absent? >> any).repeat.as(:description) >>
          newline
      end

      # @!method die
      rule(:die) { die_header >> die_body.maybe }

      # @!method ending
      rule(:ending) { any.repeat }
    end
  end
end
